import pygame
import sprites
import data
import logging
import random
import math
from util import Timer

class GameQuit(Exception):
  pass

class GameOver(Exception):
  pass

class AWinnerIsYou(Exception):
  pass

class Input(object):
  """Translate keypresses to game inputs"""
  def __init__(self, config):
    self.keys = {}
    self.controls = {}
    self.new_shot = False
    self.firing_timer = 0
    self.fire_wait_time = config.getint('weapons', 'fire_wait_time')
    for control in ('left', 'right', 'up', 'down', 'shoot'):
      keyname = config.get('controls', control)
      try:
        keyint = int(getattr(pygame, keyname))
      except ValueError():
        logging.error('You broke the config x_x')
      self.controls[keyint] = control
      self.keys[control] = keyint

  def shoot(self):
    self.new_shot = True
    self.firing_timer = self.fire_wait_time

  def update(self, ms, events):
    # Weapons cooloff
    self.firing_timer -= ms
    if self.firing_timer < 0:
      self.firing_timer = 0

    # Firing
    fire_key = self.keys['shoot']
    self.new_shot = False
    for event in events:
      if event.type == pygame.KEYDOWN and event.key == fire_key \
          and self.firing_timer == 0:
        self.shoot()

    pressed = pygame.key.get_pressed()
    self.horizontal = 0
    self.vertical = 0
    if pressed[self.keys['left']]:
      self.horizontal -= 1
    if pressed[self.keys['right']]:
      self.horizontal += 1
    if pressed[self.keys['up']]:
      self.vertical -= 1
    if pressed[self.keys['down']]:
      self.vertical += 1

    # Check if fire key is held down
    if pressed[fire_key] and self.firing_timer == 0:
      self.shoot()

class Spawner(object):
  def __init__(self, width, height, depth, config):
    self.width = width
    self.height = height
    self.depth = depth
    self.config = config
    self.rules = []

  def update(self, depth_diff):
    self.depth += depth_diff
    for rule in self.rules:
      rule.update(self.depth, self.width, self.height, self.config)
    self.rules = [rule for rule in self.rules if rule.last_spawn >= self.depth]

class SpawnPattern:
  RANDOM = 0
  LINE = 1
  LOOP = 2
  ZIG = 3
  ZAG = 4

class SpawnArea:
  LEFT = 0
  CENTER = 1
  RIGHT = 2
  ANY = 3

class SpawnRule(object):
  def __init__(self, sprite_class, first_spawn, last_spawn, seperation,
      groups, pattern=SpawnPattern.RANDOM,
      area=SpawnArea.ANY,
      shoot=sprites.Enemy.SHOOT_FWD):
    self.sprite_class = sprite_class
    self.next_spawn = first_spawn
    self.last_spawn = last_spawn
    self.seperation = seperation
    self.pattern = pattern
    self.groups = groups
    self.area = area
    self.loop_pos = None # holds position of loop start so all enemies start at same point
    self.shoot = shoot

  def get_pos(self, width, height, enemy_width, enemy_height):
    if self.area != SpawnArea.ANY:
      width /= 3

    if self.pattern == SpawnPattern.RANDOM:
      x = random.randint(width / 4, width * 3 / 4)
      if self.area != SpawnArea.ANY:
        x += width * self.area
      return ((x, height),)
    elif self.pattern in (SpawnPattern.LINE, SpawnPattern.ZIG, SpawnPattern.ZAG):
      spacingx = 20
      spacingy = 20
      offset = (width % (enemy_width + spacingx) + (enemy_width + spacingx)) / 2
      right = width
      if self.area != SpawnArea.ANY:
        offset += width * self.area
        right += width * self.area

      ret = []
      xs = range(offset, right, enemy_width + spacingx)
      for i, x in enumerate(xs):
        if self.pattern == SpawnPattern.LINE:
          y = height
        elif self.pattern == SpawnPattern.ZIG:
          y = height + (len(xs) - i) * (enemy_width + spacingy)
        else:
          y = height + i * (enemy_width + spacingy)
        ret.append((x, y))
      return ret
    elif self.pattern == SpawnPattern.LOOP:
      points = []
      radius = 100
      center_x, center_y = random.randint(int(width*1/3), int(width*2/3)), random.randint(int(height*1/3), int(height*2/3))
      start_y = random.randint(height/4, height*3/4) # starting y coordinate is same as final y coordinate
      points.append((-100, start_y)) # add starting points of path
      # add points along perimeter of circle to path
      for i in range(-2, 7): # start at -2pi/4 and end on 5pi/4 (-90 degrees -> 270 degrees)
        theta = i * math.pi / 4 # a point is added each pi/4 radians
        next_point = ( int(center_x + radius*math.cos(theta)), int(center_y - radius*math.sin(theta)) )
        points.append(next_point)
      points.append((width + 100, start_y)) # add a point off screen to the right
      return points

  def update(self, depth, width, height, config):
    shoot = self.shoot
    if depth > self.next_spawn and depth - self.next_spawn < 100:
      self.next_spawn += self.seperation
      lazy_hack = self.sprite_class((0, 0), config)
      if self.pattern == SpawnPattern.LOOP:
        if self.loop_pos == None:
          self.loop_pos = self.get_pos(width, height, lazy_hack.rect.width, lazy_hack.rect.height)
        enemy = self.sprite_class(self.loop_pos, config, shoot)
        for group in self.groups:
          group.add(enemy)
      else:
          for pos in self.get_pos(width, height, lazy_hack.rect.width, lazy_hack.rect.height):
            enemy = self.sprite_class(pos, config, shoot)
            for group in self.groups:
              group.add(enemy)

class CaptionRules:
  def __init__(self):
    self.rules = []

  def update(self, controller, width, depth):
    for rule in self.rules:
      if rule[0] <= depth:
        controller.new_caption(width, rule[2], rule[3], rule[4], rule[5])
      if rule[1] <= depth:
        controller.no_caption()
        del self.rules[0]
      return controller

  def new_rule(self, controller, width, start_depth, end_depth, text='', ypos=10, color=(245,245,245), rect_color=(64,64,64)):
    self.rules.append([start_depth, end_depth, text, ypos, color, rect_color])



class Mixer(object):
  def __init__(self, config):
    pygame.mixer.init()
    self.muted = config.getboolean('game', 'mute')
    self.current = None
    self.explosion = pygame.mixer.Sound(data.filepath('sounds', 'explosion3.wav'))
    self.explosion.set_volume(0.3)
    self.hurt = pygame.mixer.Sound(data.filepath('sounds', 'hurt.wav'))
    self.laser = pygame.mixer.Sound(data.filepath('sounds', 'laser3.wav'))
    self.laser.set_volume(0.1)
    self.inkball = pygame.mixer.Sound(data.filepath('sounds', 'laser4.wav'))
    self.inkball.set_volume(0.1)
    self.hurt = pygame.mixer.Sound(data.filepath('sounds', 'hurt.wav'))
    self.hurt.set_volume(0.75)

  def play_track(self, filename):
    if self.muted or self.current == filename:
      return
    pygame.mixer.music.load(data.filepath('music', filename))
    pygame.mixer.music.set_volume(0.5)
    pygame.mixer.music.play(-1)
    self.current = filename

  def play_sound(self, sound):
    if self.muted:
      return
    channel = pygame.mixer.find_channel()
    sound_to_play = getattr(self, sound)
    try:
      channel.play(sound_to_play)
    except:
      pass # too many sounds playing at once?

class Game(object):
  def __init__(self, controller, config, size, debug, start_depth=0):
    self.width, self.height = size
    self.nemesis_depth = config.getint('game', 'nemesis_depth')
    self.depth = start_depth
    self.scroll_speed = config.getfloat('game', 'scroll_speed')
    #self.scroll_speed = 0
    self.controller = controller
    self.player = sprites.PlayerSprite((self.width / 2, 20), config)
    self.player_layer = pygame.sprite.Group(self.player)
    self.player_bullets = pygame.sprite.Group()
    self.enemy_layer = pygame.sprite.Group()
    self.jellyfish_layer = pygame.sprite.Group()
    self.squid_layer = pygame.sprite.Group()
    self.enemy_bullets = pygame.sprite.Group()
    self.explosion_layer = pygame.sprite.Group()
    self.animations = pygame.sprite.Group(self.player)
    self.spawner = Spawner(self.width, self.height, self.depth, config)
    self.meters_per_pixel = config.getfloat('game', 'nemesis_depth_meters') / self.nemesis_depth
    self.caption_rules = CaptionRules()
    self.caption = {}
    self.config = config
    self.nemesis = None
    self.nemesis_layer = pygame.sprite.Group()
    self.checkpoints = []

    # Colours
    speech = (245,184,0)
    speech_rect = (64,64,64)
    info = (245,184,0)
    info_rect = (64,64,64)
    nem_speech = (245,184,0)
    nem_speech_rect = (64,64,64)
    centery = self.height/2 - 10
    bottom = self.height - 75
    top = 25

    # AREA 1
    self.caption_rules.new_rule(self.controller, self.width, 100, 120, \
      "OCTO-NEMESIS!!!!", centery, speech, speech_rect)
    self.caption_rules.new_rule(self.controller, self.width, 140, 160, \
      "OCTO-NEMESIS!!!!", centery, speech, speech_rect)
    self.caption_rules.new_rule(self.controller, self.width, 180, 350, \
      "OCTO-NEMESIS!!!!", centery, speech, speech_rect)
    self.caption_rules.new_rule(self.controller, self.width, 500, 700, \
      "He's down here somewhere...", centery, speech, speech_rect)
    self.caption_rules.new_rule(self.controller, self.width, 800, 1000, \
      "I must find what he's up to!", centery, speech, speech_rect)
    self.caption_rules.new_rule(self.controller, self.width, 1200, 1300, \
      "ENTERING AREA 1", bottom, info, info_rect)
    self.checkpoints.append(1100)

    #self.caption_rules.new_rule(self.controller, self.width, 500, 700, \
    #  "Weapon Upgraded!", self.height-150, (245,184,0), (64,64,64))

    area_start = 1100
    self.spawner.rules.append(
        SpawnRule(sprites.Jellyfish, area_start, area_start + 5000, 100,
          (self.enemy_layer, self.jellyfish_layer, self.animations))
    )

    self.caption_rules.new_rule(self.controller, self.width, area_start + 1000, area_start + 1200, \
      "Jellyfish creep me out", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 1500, area_start + 1700, \
      "Must keep calm...", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 1900, area_start + 2200, \
      "The key is to stay in Ctrl!", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 2500, area_start + 2700, \
      "Ha ha ha", bottom, speech, speech_rect)

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 3000, area_start + 3000, 100,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.RANDOM,
        )
    )

    self.caption_rules.new_rule(self.controller, self.width, area_start + 3020, area_start + 3220, \
      "EEK! A squid!", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 3500, area_start + 3700, \
      "That was close!", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 4000, area_start + 4200, \
      "One touch could be deadly", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 4300, area_start + 4500, \
      "And they can shoot ink", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 4700, area_start + 4900, \
      "Deadly, deadly ink", bottom, speech, speech_rect)

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 5100, area_start + 5300, 200,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.RANDOM, area=SpawnArea.LEFT)
    )
    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 5200, area_start + 5300, 200,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.RANDOM, area=SpawnArea.CENTER)
    )
    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 5100, area_start + 5300, 200,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.RANDOM, area=SpawnArea.RIGHT)
    )

    self.caption_rules.new_rule(self.controller, self.width, area_start + 6000, area_start + 6200, \
        "It's like mom always said:", bottom, speech, speech_rect)
    self.caption_rules.new_rule(self.controller, self.width, area_start + 6400, area_start + 6600, \
      '"Never trust a cephalopod"', bottom, speech, speech_rect)

    self.spawner.rules.append(
        SpawnRule(sprites.Jellyfish, area_start + 6400, area_start + 8000, 200,
          (self.enemy_layer, self.jellyfish_layer, self.animations))
    )

    self.caption_rules.new_rule(self.controller, self.width, area_start + 6700, area_start + 6900, \
      "Bet Octo himself sent 'em", bottom, speech, speech_rect)

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 6800, area_start + 6800, 200,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.LINE, area=SpawnArea.RIGHT)
    )

    self.caption_rules.new_rule(self.controller, self.width, area_start + 7100, area_start + 7300, \
      "He must know I'm coming", bottom, speech, speech_rect)

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 7200, area_start + 7400, 200,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.LINE, area=SpawnArea.LEFT)
    )

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 7300, area_start + 7300, 200,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.LINE, area=SpawnArea.CENTER)
    )

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 7300, area_start + 7800, 500,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.RANDOM, area=SpawnArea.RIGHT)
    )

    self.spawner.rules.append(
        SpawnRule(sprites.Jellyfish, area_start + 8000, area_start + 8900, 300,
          (self.enemy_layer, self.jellyfish_layer, self.animations))
    )

    self.caption_rules.new_rule(self.controller, self.width, area_start + 8000, area_start + 8200, \
      "Phew", bottom, info, info_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 8300, area_start + 8500, \
      "Good job I bought my nanobots", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 8700, area_start + 8900, \
      "Could save my life down here", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 9100, area_start + 9300, \
      "If only I had more batteries", top, speech, speech_rect)

    area_start += 9700
    self.checkpoints.append(area_start)
    self.caption_rules.new_rule(self.controller, self.width, area_start, area_start + 200, \
      "ENTERING AREA 2", bottom, info, info_rect)
    self.spawner.rules.append(
        SpawnRule(sprites.Jellyfish, area_start + 100, area_start + 3000, 300,
          (self.enemy_layer, self.jellyfish_layer, self.animations))
    )

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 200, area_start + 400, 300,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.LINE)
    )

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 500, area_start + 3000, 300,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.LINE, area = SpawnArea.LEFT)
    )

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 500, area_start + 3000, 300,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.LINE, area = SpawnArea.RIGHT)
    )

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 900, area_start + 3000, 600,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.LINE, area = SpawnArea.CENTER,
          shoot = sprites.Enemy.AIM)
    )

    self.caption_rules.new_rule(self.controller, self.width, area_start + 3100, area_start + 3300, \
      "He must've been training 'em", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 3500, area_start + 3700, \
      "But for WHAT?", bottom, speech, speech_rect)

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 3500, area_start + 3500, 100,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.LINE, area=SpawnArea.LEFT,
          shoot = sprites.Enemy.AIM)
    )

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 3700, area_start + 3700, 100,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.LINE, area=SpawnArea.RIGHT,
          shoot = sprites.Enemy.AIM)
    )

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 2900, area_start + 2900, 100,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.ZIG, area=SpawnArea.CENTER,
          shoot = sprites.Enemy.AIM)
    )
    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 4100, area_start + 4100, 100,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.ZAG, area=SpawnArea.ANY,
          shoot = sprites.Enemy.AIM)
    )

    self.caption_rules.new_rule(self.controller, self.width, 15000, 15200, \
      "WEAPONS UPGRADED", bottom, info, info_rect)


    # LOOPING SEA URCHINS
    self.spawner.rules.append(
        SpawnRule(sprites.Urchin, area_start +  4300, area_start +4600, 50,
          (self.enemy_layer, self.animations),
          pattern = SpawnPattern.LOOP)
    )

    # LOOPING AND SHOOTING SQUIDS
    self.spawner.rules.append(
        SpawnRule(sprites.Squid2, area_start + 5000, area_start + 5200, 100,
          (self.enemy_layer, self.animations),
          pattern = SpawnPattern.LOOP)
    )

    self.caption_rules.new_rule(self.controller, self.width, area_start + 5600, area_start + 5800, \
      "Muahahaha", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 5900, area_start + 6000, \
      "I will destroy you all!", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 6200, area_start + 6400, \
      "Ahem", bottom, speech, speech_rect)

    self.spawner.rules.append(
        SpawnRule(sprites.Urchin, area_start +  6400, area_start + 6600, 50,
          (self.enemy_layer, self.animations),
          pattern = SpawnPattern.LOOP)
    )

    self.spawner.rules.append(
        SpawnRule(sprites.Urchin, area_start +  6600, area_start + 6800, 100,
          (self.enemy_layer, self.animations),
          pattern = SpawnPattern.LOOP)
    )

    self.spawner.rules.append(
        SpawnRule(sprites.Urchin, area_start +  6800, area_start + 7000, 100,
          (self.enemy_layer, self.animations),
          pattern = SpawnPattern.LOOP)
    )

    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 6500, area_start + 7100, 300,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.ZIG, area=SpawnArea.RIGHT,
          shoot = sprites.Enemy.AIM)
    )
    self.spawner.rules.append(
        SpawnRule(sprites.Squid, area_start + 6500, area_start + 7100, 300,
          (self.enemy_layer, self.squid_layer, self.animations),
          pattern = SpawnPattern.ZAG, area=SpawnArea.RIGHT,
          shoot = sprites.Enemy.AIM)
    )

    for i in range(0, 600, 100):
      self.spawner.rules.append(
          SpawnRule(sprites.Squid2, area_start + 7300 + i, area_start + 7300 + i, 100,
            (self.enemy_layer, self.animations),
            pattern = SpawnPattern.LOOP)
      )

    self.spawner.rules.append(
      SpawnRule(sprites.Urchin, area_start +  8000, area_start + 9200, 400,
        (self.enemy_layer, self.animations),
        pattern = SpawnPattern.LOOP)
    )

    for i in range(0, 1000, 100):
      self.spawner.rules.append(
        SpawnRule(sprites.Jellyfish, area_start + 7800 + i, area_start + 7800, 100,
        (self.enemy_layer, self.jellyfish_layer, self.animations), area=SpawnArea.LEFT)
      )

      self.spawner.rules.append(
        SpawnRule(sprites.Jellyfish, area_start + 7800 + i, area_start + 7800, 100,
        (self.enemy_layer, self.jellyfish_layer, self.animations), area=SpawnArea.RIGHT)
      )

      self.spawner.rules.append(
          SpawnRule(sprites.Squid2, area_start + 7800 + i, area_start + 7800 + i + 100, 100,
            (self.enemy_layer, self.animations),
            pattern = SpawnPattern.LOOP)
      )

    self.spawner.rules.append(
      SpawnRule(sprites.Urchin, area_start +  9500, area_start + 10000, 100,
        (self.enemy_layer, self.animations),
        pattern = SpawnPattern.LINE)
      )

    area_start += 10100
    self.caption_rules.new_rule(self.controller, self.width, area_start, area_start + 200, \
      "Oh hi", centery, nem_speech, nem_speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 400, area_start + 600, \
      "You fell right into my trap", centery, nem_speech, nem_speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 800, area_start + 1000, \
      "While you were sightseeing", centery, nem_speech, nem_speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 1200, area_start + 1400, \
      "I reprogrammed the game", centery, nem_speech, nem_speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 1600, area_start + 1800, \
      "I am now invincible", centery, nem_speech, nem_speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 2000, area_start + 2200, \
      "And I have all the weapons", centery, nem_speech, nem_speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 2400, area_start + 2600, \
      "Also we invaded the surface", centery, nem_speech, nem_speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 2800, area_start + 3000, \
      "And enslaved your species", centery, nem_speech, nem_speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 3500, area_start + 3700, \
      "Oh dear...", bottom, speech, speech_rect)

    self.caption_rules.new_rule(self.controller, self.width, area_start + 4000, area_start + 4200, \
      "NEMESIS APPROACHING", bottom, speech, speech_rect)
    self.checkpoints.append(area_start + 4000)

    self.inputs = Input(config)
    self.debug = debug
    self.animation_timer = Timer(100, self.animate)


  def summon_nemesis(self):
    self.nemesis = sprites.Nemesis([
      (1000, 1000), (20, 20), (250, 20), (250, 200), (20, 200)
      ], self.config, self.width, self.height)
    self.animations.add(self.nemesis)
    self.nemesis_layer.add(self.nemesis)

  def animate(self):
    for animation in self.animations:
      animation.animate()

  def cleanup(self):
    for bullet in self.player_bullets:
      if bullet.rect.top > self.height:
        logging.debug('Bye bullet')
        bullet.kill()
    for enemy in self.enemy_layer:
      if enemy.rect.bottom < 0:
        logging.debug('Bye enemy')
        enemy.kill()
      elif enemy.rect.x > self.width:
        logging.info('EXIT STAGE RIGHT')
        enemy.kill()
    for bullet in self.enemy_bullets:
      if bullet.rect.bottom < 0 or bullet.rect.top > self.height or \
          bullet.rect.right < 0 or bullet.rect.left > self.width:
        logging.debug('Bye bullet')
        bullet.kill()

    for explosion in self.explosion_layer:
      if explosion.finished == True:
        explosion.kill()

  def collide_enemies(self, explosion_group):
    for player_bullet in self.player_bullets:
      if self.nemesis:
        self.nemesis.collide(player_bullet, explosion_group, self.animations, self.controller.mixer)
    if self.nemesis and not self.nemesis.dead and not self.player.invulnerable and not self.player.dead and (self.nemesis.arm_collides(self.player.colliderect) or self.nemesis.head.rect.colliderect(self.player.colliderect)):
        logging.info('COLLIDE')
        self.controller.mixer.play_sound('explosion')
        self.player.die((explosion_group, self.animations))

    for enemy in self.enemy_layer:
      dead = False
      for player_bullet in self.player_bullets:
        if enemy.colliderect.colliderect(player_bullet.rect):
          player_bullet.kill()
          enemy.health -= player_bullet.power
          if enemy.health < 0:
            enemy.kill()
            dead = True
            logging.info('BOOM')
            self.controller.mixer.play_sound('explosion')
            pos = (enemy.rect.center[0], enemy.rect.top) # x=center of rect, y=top of rect
            explosion = sprites.Explosion(pos)
            explosion_group.add(explosion)
            self.animations.add(explosion)
            break
      if not dead and not self.player.dead and not self.player.invulnerable and enemy.colliderect.colliderect(self.player.colliderect):
        if enemy in self.jellyfish_layer:
          self.player.stun()
        else:
          logging.info('COLLIDE')
          self.controller.mixer.play_sound('explosion')
          self.player.die((explosion_group, self.animations))

    for bullet in self.enemy_bullets:
      if bullet.rect.colliderect(self.player.colliderect) and not self.player.invulnerable:
        self.controller.mixer.play_sound('explosion')
        self.player.die((explosion_group, self.animations))
        bullet.kill()

  def run(self, ms, events):
    # Enemies are not moving downwards with the player
    distance = ms * self.scroll_speed

    if self.nemesis:
      try:
        self.nemesis.update(ms, self.player, self.controller.mixer, self.enemy_bullets, [self.enemy_layer, self.animations, self.jellyfish_layer])
      except sprites.Nemesis.NOOOOOooo:
        raise AWinnerIsYou
    for obj in self.enemy_layer:
      if not obj.has_path:
        # Movement relative to the diver
        obj.rect.top -= distance
        obj.colliderect.center = obj.rect.center
    self.depth += distance
    for checkpoint in self.checkpoints:
      if checkpoint < self.depth:
        self.controller.level = checkpoint
    self.inputs.update(ms, events)
    self.spawner.update(distance)
    # squids shoot
    for enemy in self.enemy_layer:
      enemy.update(ms, self.player, self.controller.mixer, self.enemy_bullets, [self.enemy_layer, self.animations])
    self.enemy_bullets.update(ms)

    self.player.update(ms, self.depth)
    self.player_bullets.update(ms)
    self.player.moveright(self.inputs.horizontal * ms)
    self.player.movedown(self.inputs.vertical * ms)
    if self.inputs.new_shot and not self.player.stunned and not self.player.dead:
      self.controller.mixer.play_sound('laser')
      self.player_bullets.add(self.player.shoot())

    # TODO add collision rects
    if self.player.rect.left < 0:
      self.player.rect.left = 0
    elif self.player.rect.right > self.width:
      self.player.rect.right = self.width
    if self.player.rect.top < 0:
      self.player.rect.top = 0
    elif self.player.rect.bottom > self.height:
      self.player.rect.bottom = self.height

    self.collide_enemies(self.explosion_layer)
    self.cleanup()
    if self.player.lives <= 0:
      raise GameOver()

    self.caption_rules.update(self.controller, self.width, self.depth)

    self.animation_timer.update(ms)

    if self.depth > self.nemesis_depth and self.nemesis is None:
      self.summon_nemesis()

  def draw_bg(self, surface):
    # TODO: make this use tiles, and add some parallax
    tile_size = 100
    offset = int(self.depth % tile_size)
    width, height = surface.get_size()
    for i in range(0 - offset, height, tile_size):
      pygame.draw.rect(surface, (0, 0, 120), pygame.Rect(0, i, width, tile_size / 2))
      pygame.draw.rect(surface, (0, 0, 150), pygame.Rect(0, i + tile_size / 2, width, tile_size / 2))

  def draw_widgets(self, surface):
    # Lives
    offsetx = 10
    offsety = 10
    width = 12
    height = 12
    spacing = 10
    color = (100, 100, 255)
    for i in xrange(self.player.lives):
      pygame.draw.rect(surface, color, pygame.Rect(offsetx, offsety, width, height))
      offsetx += width + spacing

    text = self.controller.fonts['ps16'].render('{0}m'.format(int(self.depth * self.meters_per_pixel)),
        False, (100, 100, 255))
    text_width = text.get_width()
    surface.blit(text, (self.width - text_width - 10, 10))

    if self.debug:
      text = self.controller.fonts['ps16'].render('{0} Enemies'.format(len(self.enemy_layer)),
        False, (100, 100, 255))
      text_width = text.get_width()
      surface.blit(text, (self.width - text_width - 10, self.height-75))

      text = self.controller.fonts['ps16'].render('{0} Enemy Bullets'.format(len(self.enemy_bullets)),
        False, (100, 100, 255))
      text_width = text.get_width()
      surface.blit(text, (self.width - text_width - 10, self.height-50))

      text = self.controller.fonts['ps16'].render('{0} Player Bullets'.format(len(self.player_bullets)),
        False, (100, 100, 255))
      text_width = text.get_width()
      surface.blit(text, (self.width - text_width - 10, self.height-25))

  def draw(self, surface):
    self.draw_bg(surface)
    self.player_bullets.draw(surface)
    self.enemy_bullets.draw(surface)
    self.enemy_layer.draw(surface)
    if self.nemesis and not self.nemesis.dead:
      self.nemesis_layer.draw(surface)
    if not self.player.dead:
      self.player_layer.draw(surface)
    self.explosion_layer.draw(surface)
    self.draw_widgets(surface)
    self.controller.draw_caption(surface)

    if self.debug:
      for enemy in self.enemy_layer:
        pygame.draw.rect(surface, (255, 0, 0), enemy.colliderect, 1)
      pygame.draw.rect(surface, (255, 0, 0), self.player.colliderect, 1)

      if self.nemesis:
        pygame.draw.rect(surface, (255, 0, 0), self.nemesis.rect, 1)
        pygame.draw.rect(surface, (255, 0, 0), self.nemesis.head.rect, 1)

class Menu(object):
  def __init__(self, controller, config):
    self.controller = controller
    self.surface = pygame.display.get_surface()
    self.image = pygame.image.load(data.filepath('images', 'menu.png'))
    self.width = self.surface.get_width()
    self.height = self.surface.get_height()
    self.surface.fill((64,64,255))
    self.player = sprites.PlayerSprite((self.width / 2, 20), config)
    self.player_layer = sprites.Group(self.player)
    self.animations = sprites.Group(self.player)
    self.animation_timer = Timer(100, self.animate)

  def animate(self):
    for animation in self.animations:
      animation.animate()


  def run(self, dt, events):
    for event in events:
      if event.type == pygame.KEYUP and event.key == pygame.K_SPACE:
        self.controller.start_game()
        self.menu = False
        return 
    self.animation_timer.update(dt)

  def draw(self, surface):
    #self.surface.fill((64,64,255))
    surface.blit(self.image, (0,0))
    self.player_layer.draw(self.surface)

class GameOverScreen(object):
  def __init__(self, controller, game, winner=False):
    self.controller = controller
    self.game = game
    self.winner = winner

  def run(self, dt, events):
    self.game.animation_timer.update(dt)
    self.game.cleanup()
    for event in events:
      if event.type == pygame.KEYUP and event.key == pygame.K_SPACE:
        self.controller.start_game()
        return

  def draw(self, surface):
    self.game.draw(surface)
    width, height = surface.get_size()
    if self.winner:
      title = self.controller.fonts['ps22'].render('YOU WIN!', False, (255, 255, 255))
      subtitle = self.controller.fonts['ps16'].render('But due to your hesitation', False, (255, 255, 255))
      subtitle2 = self.controller.fonts['ps16'].render('humanity was destroyed.', False, (255, 255, 255))
      spacing = 30
      title_width, title_height = title.get_size()
      subtitle_width, subtitle_height = subtitle.get_size()
      subtitle2_width, subtitle2_height = subtitle2.get_size()
      x = (width - title_width) / 2
      y = (height - title_height - spacing - subtitle_height - subtitle2_height) / 2
      surface.blit(title, (x, y))

      x = (width - subtitle_width) / 2
      y += title_height + spacing
      surface.blit(subtitle, (x, y))

      x = (width - subtitle2_width) / 2
      y += subtitle_height + spacing
      surface.blit(subtitle2, (x, y))
    else:
      title = self.controller.fonts['ps22'].render('YOU LOSE!', False, (255, 255, 255))
      subtitle = self.controller.fonts['ps16'].render('Press time to activate', False, (255, 255, 255))
      subtitle2 = self.controller.fonts['ps16'].render('space machine', False, (255, 255, 255))
    spacing = 30
    title_width, title_height = title.get_size()
    subtitle_width, subtitle_height = subtitle.get_size()
    subtitle2_width, subtitle2_height = subtitle2.get_size()
    x = (width - title_width) / 2
    y = (height - title_height - spacing - subtitle_height - subtitle2_height) / 2
    surface.blit(title, (x, y))

    x = (width - subtitle_width) / 2
    y += title_height + spacing
    surface.blit(subtitle, (x, y))

    x = (width - subtitle2_width) / 2
    y += subtitle_height + spacing
    surface.blit(subtitle2, (x, y))

class Controller(object):
  def __init__(self, config, surface, debug, level):
    self.config = config
    self.scenes = {
        'menu': Menu(self, config),
    }
    self.surface = surface
    self.debug = debug
    self.level = level
    self.fonts = {
      'ps16' : pygame.font.Font(data.filepath('press-start-2p', 'PressStart2P.ttf'), 16),
      'ps22' : pygame.font.Font(data.filepath('press-start-2p', 'PressStart2P.ttf'), 22)
    }
    self.mixer = Mixer(config)
    self.mixer.play_track('menu.ogg')
    self.caption = {}

  def new_caption(self, width, text='', ypos=10, color=(245,245,245), rect_color=(64,64,64)):
    self.caption['text'] = text
    self.caption['color'] = color
    text_width, text_height = self.fonts['ps16'].size(text)
    margin = text_height / 2
    text_width = text_width + 2*margin
    text_height = text_height + 2*margin
    rect_y = ypos - margin
    rect_x = width/2 - text_width/2
    self.caption['ypos'] = ypos
    self.caption['xpos'] = width/2 - text_width/2 + margin
    self.caption['rect'] = (rect_x, rect_y, text_width, text_height)
    self.caption['rect_color'] = rect_color
    self.caption['rendered'] = self.fonts['ps16'].render(text, False, self.caption['color'])

  def no_caption(self):
    self.caption = {}

  def draw_caption(self, surface):
    if self.caption:
      pygame.draw.rect(surface, self.caption['rect_color'], self.caption['rect'])
      pygame.draw.rect(surface, self.caption['color'], ( self.caption['rect'][0]+2, self.caption['rect'][1]+2, self.caption['rect'][2]-4, self.caption['rect'][3]-4) , 1)
      surface.blit(self.caption['rendered'], (self.caption['xpos'], self.caption['ypos']))

  def start_game(self):
    game = Game(self, self.config, self.surface.get_size(), self.debug, self.level)
    self.scenes['game'] = game
    self.active = game
    self.mixer.play_track('game.ogg')
    self.no_caption()

  def change_scene(self, newscene):
    self.active = self.scenes[newscene]

  def poll_input(self):
    for event in pygame.event.get():
      if event.type == pygame.QUIT:
        raise GameQuit()
      else:
        yield event

  def loop(self):
    #pygame.mixer.pre_init(22050, -16, 2, 1024) # setup mixer to avoid sound lag
    pygame.init()                      #initialize pygame
    max_fps = self.config.getint('game', 'max_fps')
    clock = pygame.time.Clock()
    self.menu = True
    self.active = Menu(self, self.config)

    # DELETE THIS \/
    self.new_caption(pygame.display.get_surface().get_width(), 'Press Space to Start', 350, (245,184,0), (66,107,212))
    while True:
      ms = clock.tick(max_fps)
      events = list(self.poll_input())
      try:
        self.active.run(ms, events)
      except GameOver:
        self.active = GameOverScreen(self, self.active)
      except AWinnerIsYou:
        self.active = GameOverScreen(self, self.active, True)
      self.active.draw(self.surface)
      self.draw_caption(self.surface)
      pygame.display.flip()
